"""Console script for terragraph."""
import click

from terragraph.core import HighlightingMode, create_highlighted_svg, from_file


@click.group()
def terragraph_cli() -> click.BaseCommand:
    """Console script for terragraph."""


@terragraph_cli.command()
@click.option(
    "--file-name", required=True, type=click.Path(exists=True, dir_okay=False)
)
@click.option("--node-name", required=True, help="Name of the node to highlight")
@click.option(
    "--mode",
    type=click.Choice([e.value for e in HighlightingMode]),
    default=HighlightingMode.PRECEDING.value,
    help="Select highlighting mode",
)
def highlight(file_name: str, node_name: str, mode: HighlightingMode) -> int:
    """
    Highlights a node and its edges
    """
    create_highlighted_svg(file_name, node_name, mode=HighlightingMode(mode))
    return 0


@terragraph_cli.command()
@click.option(
    "--file-name", required=True, type=click.Path(exists=True, dir_okay=False)
)
def show_nodes(file_name: str) -> int:
    """
    Lists nodes in the terraform graph
    """
    print(f"file is: {file_name}")
    tfg = from_file(file_name)
    for node in tfg.get_node_names():
        print(node)

    return 0



"""
Test cases for the CLI
"""
import pytest
from click.testing import CliRunner
from terragraph.cli import terragraph_cli


@pytest.fixture
def runner():
    """
    Creates a CLIRunner
    """
    return CliRunner()


@pytest.fixture()
def runner_temp_file(tmp_path):
    """
    Creates a temp file for the runner to use with a terraform graph in it
    """
    # Create a temporary file for testing
    test_file = tmp_path / "graph.dot"
    with open(test_file, "w", encoding="utf-8") as f:
        f.write(
            r"""
digraph {
	compound = "true"
	newrank = "true"
	subgraph "root" {
		"[root] module.mod1.random_pet.this (expand)" [label = "module.mod1.random_pet.this", shape = "box"]
		"[root] module.mod1.random_pet.this2 (expand)" [label = "module.mod1.random_pet.this2", shape = "box"]
		"[root] module.mod1.time_sleep.this (expand)" [label = "module.mod1.time_sleep.this", shape = "box"]
		"[root] module.mod2.random_pet.this (expand)" [label = "module.mod2.random_pet.this", shape = "box"]
		"[root] provider[\"registry.terraform.io/hashicorp/random\"]" [label = "provider[\"registry.terraform.io/hashicorp/random\"]", shape = "diamond"]
		"[root] provider[\"registry.terraform.io/hashicorp/time\"]" [label = "provider[\"registry.terraform.io/hashicorp/time\"]", shape = "diamond"]
		"[root] var.pet_length" [label = "var.pet_length", shape = "note"]
		"[root] module.mod1 (close)" -> "[root] module.mod1.output.pet (expand)"
		"[root] module.mod1 (close)" -> "[root] module.mod1.output.pet_length (expand)"
		"[root] module.mod1 (close)" -> "[root] module.mod1.random_pet.this2 (expand)"
		"[root] module.mod1.output.pet (expand)" -> "[root] module.mod1.random_pet.this (expand)"
		"[root] module.mod1.output.pet_length (expand)" -> "[root] module.mod1.random_pet.this (expand)"
		"[root] module.mod1.random_pet.this (expand)" -> "[root] module.mod1.var.pet_length (expand)"
		"[root] module.mod1.random_pet.this (expand)" -> "[root] provider[\"registry.terraform.io/hashicorp/random\"]"
		"[root] module.mod1.random_pet.this2 (expand)" -> "[root] module.mod1.time_sleep.this (expand)"
		"[root] module.mod1.random_pet.this2 (expand)" -> "[root] module.mod1.var.pet_length (expand)"
		"[root] module.mod1.random_pet.this2 (expand)" -> "[root] provider[\"registry.terraform.io/hashicorp/random\"]"
		"[root] module.mod1.time_sleep.this (expand)" -> "[root] module.mod1 (expand)"
		"[root] module.mod1.time_sleep.this (expand)" -> "[root] provider[\"registry.terraform.io/hashicorp/time\"]"
		"[root] module.mod1.var.pet_length (expand)" -> "[root] module.mod1 (expand)"
		"[root] module.mod1.var.pet_length (expand)" -> "[root] var.pet_length"
		"[root] module.mod2 (close)" -> "[root] module.mod2.random_pet.this (expand)"
		"[root] module.mod2.random_pet.this (expand)" -> "[root] module.mod2.var.pet_length (expand)"
		"[root] module.mod2.var.pet_length (expand)" -> "[root] module.mod1.output.pet (expand)"
		"[root] module.mod2.var.pet_length (expand)" -> "[root] module.mod2 (expand)"
		"[root] provider[\"registry.terraform.io/hashicorp/random\"] (close)" -> "[root] module.mod1.random_pet.this2 (expand)"
		"[root] provider[\"registry.terraform.io/hashicorp/random\"] (close)" -> "[root] module.mod2.random_pet.this (expand)"
		"[root] provider[\"registry.terraform.io/hashicorp/time\"] (close)" -> "[root] module.mod1.time_sleep.this (expand)"
		"[root] root" -> "[root] module.mod1 (close)"
		"[root] root" -> "[root] module.mod2 (close)"
		"[root] root" -> "[root] provider[\"registry.terraform.io/hashicorp/random\"] (close)"
		"[root] root" -> "[root] provider[\"registry.terraform.io/hashicorp/time\"] (close)"
	}
}
"""
        )
        f.flush()
        yield test_file


def test_highlight_command(runner, runner_temp_file):
    """
    Will test the highlight command with its default inputs
    """
    # Create a temporary file for testing
    input_file_path = runner_temp_file.absolute()
    output_file_path = f"{input_file_path}.svg"
    result = runner.invoke(
        terragraph_cli,
        ["highlight", "--file-name", input_file_path, "--node-name", '"[root] module.mod1.random_pet.this2 (expand)"']
    )

    assert result.exit_code == 0
    assert f"Colored node SVG file generated: {output_file_path}" in result.output


def test_show_nodes_command(runner, runner_temp_file):
    """
    test valid show-nodes command
    """
    input_file_path = runner_temp_file.absolute()
    result = runner.invoke(
        terragraph_cli,
        ["show-nodes", "--file-name", input_file_path]
    )

    expected_nodes = r"""
"[root] module.mod1.random_pet.this (expand)"
"[root] module.mod1.random_pet.this2 (expand)"
"[root] module.mod1.time_sleep.this (expand)"
"[root] module.mod2.random_pet.this (expand)"
"[root] provider[\"registry.terraform.io/hashicorp/random\"]"
"[root] provider[\"registry.terraform.io/hashicorp/time\"]"
"""

    assert result.exit_code == 0
    assert f"file is: {input_file_path}" in result.output
    assert expected_nodes in result.output  # Check if actual nodes are listed


def test_invalid_highlight_mode(runner, runner_temp_file):
    """
    Tests that the command fails for invalid highlight modes
    """
    result = runner.invoke(
        terragraph_cli,
        ["highlight", "--file-name", runner_temp_file.absolute(), "--node-name", "test-node", "--mode", "invalid_mode"]
    )

    assert result.exit_code != 0
    assert "Invalid value for '--mode'" in result.output



"""Top-level package for terragraph."""

__author__ = "Chris Doyle"
__email__ = "uk.chris.doyle+cdsre@gmail.com"
__version__ = "0.3.3"

from .core import Terragraph, from_file, create_highlighted_svg, HighlightingMode



"""
The core of the terragraph module
"""
from enum import Enum
import pydot  # type: ignore


class HighlightingMode(Enum):
    """
    An Enum for managing the highlighting modes for edges
    """

    ALL = "all"
    PRECEDING = "preceding"
    SUCCESSOR = "successor"


class Terragraph:
    """
    A class that will parse the output from a `terraform graph` command and can highlight a node and its associated edges.
    It can highlight preceeding edges, or succesor edges. It can also highlight both allowing the full dependency tree
    top and bottom for a given node.
    """

    DEFAULT_HIGHLIGHTING_MODE = HighlightingMode.PRECEDING

    def __init__(
        self,
        dot_data: str,
        subgraph_name: str = '"root"',
        highlighting_mode: HighlightingMode = DEFAULT_HIGHLIGHTING_MODE,
    ):
        self.__graph = pydot.graph_from_dot_data(dot_data)[0]
        self.tf_graph = self.__graph.get_subgraph(subgraph_name)[0]
        self.highlight_mode = highlighting_mode

    def __get_preceding_edges(self, edges: list[pydot.Edge]) -> list[pydot.Edge]:
        destination_edge_names: list[pydot.Edge] = [
            edge.get_destination() for edge in edges
        ]

        if destination_edge_names:
            destination_edges: list[pydot.Edge] = [
                edge
                for edge in self.tf_graph.get_edges()
                if edge.get_source() in destination_edge_names
            ]
            return destination_edges + self.__get_preceding_edges(destination_edges)
        return []

    def __get_successor_edges(self, edges: list[pydot.Edge]) -> list[pydot.Edge]:
        successor_edge_names: list[pydot.Edge] = [edge.get_source() for edge in edges]

        if successor_edge_names:
            successor_edges: list[pydot.Edge] = [
                edge
                for edge in self.tf_graph.get_edges()
                if edge.get_destination() in successor_edge_names
            ]
            return successor_edges + self.__get_successor_edges(successor_edges)
        return []

    def highlight_node(self, node_name: str, color: str = "red") -> None:
        """
        Highlights a node and its given edges
        :param node_name: The name of the node to highlight from
        :param color: The color to highlight the node
        :return: None
        """
        node = self.tf_graph.get_node(node_name)[0]
        node.set_color(color)
        self.highlight_node_edges(node_name)

    @staticmethod
    def highlight_edges(edges: list[pydot.Edge], color: str = "red") -> None:
        """
        highlights a list of edges with the color name
        :param edges: A list of edge objects to be highlighted
        :param color: The color to highlight the edges
        :return: None
        """
        for edge in edges:
            edge.set_color(color)

    def __get_all_preceding_edges(self, node_name: str) -> list[pydot.Edge]:
        node_preceding_edges = [
            edge for edge in self.tf_graph.get_edges() if edge.get_source() == node_name
        ]
        preceding_edges = self.__get_preceding_edges(node_preceding_edges)
        return node_preceding_edges + preceding_edges

    def __get_all_successor_edges(self, node_name: str) -> list[pydot.Edge]:
        node_successor_edges = [
            edge
            for edge in self.tf_graph.get_edges()
            if edge.get_destination() == node_name
        ]
        successor_edges = self.__get_successor_edges(node_successor_edges)
        return node_successor_edges + successor_edges

    def highlight_node_edges(self, node_name: str) -> None:
        """
        Takes a node name and will highlight the node and its edges based on the self.highlight_mode value
        :param node_name: The name of the node to highlight and the edges from it.
        :return: None
        """
        edges: list[pydot.Edge] = []
        if self.highlight_mode in [HighlightingMode.PRECEDING, HighlightingMode.ALL]:
            edges += self.__get_all_preceding_edges(node_name)
        if self.highlight_mode in [HighlightingMode.SUCCESSOR, HighlightingMode.ALL]:
            edges += self.__get_all_successor_edges(node_name)
        self.highlight_edges(edges)

    def get_node_names(self) -> list[str]:
        """
        Gets a list of all the nodes in the terraform_graph
        :return: list of node names
        """
        return [node.get_name() for node in self.tf_graph.get_nodes()]

    def write_svg(self, file_name: str) -> None:
        """
        Write an SVG of the current graph state
        :param file_name: output file name
        :return: None
        """
        self.__graph.write(file_name, format="svg")


def create_highlighted_svg(
    dot_file_name: str,
    highlighted_node_name: str,
    mode: HighlightingMode = Terragraph.DEFAULT_HIGHLIGHTING_MODE,
) -> None:
    """
    Will create a highlighted representation of the graph under the same path as the dot_file_name but suffixed with .svg

    :param dot_file_name: The name/path to a file containing a terraform graph output
    :param highlighted_node_name: The node name to highlight in the graph and its edges.
    :param mode: An Enum indicating which highlighting mode to use
    :return: This does not return anything as it will create the SVG in the file system.
    """
    terragraph = from_file(dot_file_name, mode)
    terragraph.highlight_node(highlighted_node_name)

    # Output the SVG file
    output_file_path = f"{dot_file_name}.svg"
    terragraph.write_svg(output_file_path)

    print(f"Colored node SVG file generated: {output_file_path}")


def from_file(filename: str, mode: HighlightingMode = Terragraph.DEFAULT_HIGHLIGHTING_MODE) -> Terragraph:
    "Takes a file name containing a 'terraform graph' output and returns a Terragraph object of it"
    with open(filename, encoding="utf-8") as dot_file:
        dot_input = dot_file.read()
        return Terragraph(dot_input, highlighting_mode=mode)
